
from .python_object_namespace import PythonObjectNamespace
from .dumped_object_namespace import DumpedObjectNamespace


class CodeObjectNamespace(DumpedObjectNamespace):

    def __init__(self, name, module, filename, lineno, type, attributes={}):
        super(CodeObjectNamespace, self).__init__(type, u'code', attributes, self.METHODS)

        self.name = name
        self.module = module
        self.filename = filename
        self.lineno = lineno

    def name(self, args):
        return PythonObjectNamespace(self.name)

    def module(self, args):
        return PythonObjectNamespace(self.module)

    def filename(self, args):
        return PythonObjectNamespace(self.filename)

    def lineno(self, args):
        return PythonObjectNamespace(self.lineno)

    def to_dict(self):
        return {
            u'@namespace': self.__class__.__name__,
            u'@common_type': self.common_type,
            u'@original_type': self.type,
            u'@attributes': self.get_attributes_dict(),
            u'@value': {
                u'name': self.name,
                u'module': self.module,
                u'filename': self.filename,
                u'lineno': self.lineno
            }
        }

    def to_simple_dict(self):
        return '%s @ %s' % (self.name, self.module)

    METHODS = (name, module, filename, lineno)



from rook.processor.paths.arithmetic_path import ArithmeticPath


class IfCondition(object):
    NAME = 'if'

    def __init__(self, configuration, factory):
        path_config = {
            'name': 'calc',
            'path': configuration['path']
        }

        self.path = ArithmeticPath(path_config, factory)

    def evaluate(self, namespace, extracted):
        obj = self.path.read_from(namespace).obj

        return isinstance(obj, bool) and obj



class ActionRunProcessor(object):
    NAME = 'script'

    def __init__(self, configuration, processor_factory):
        self.processor = processor_factory.get_processor(configuration['operations'])

        if 'post_operations' in configuration:
            self.post_processor = processor_factory.get_processor(configuration['post_operations'])
        else:
            self.post_processor = None

    def execute(self, aug_id, namespace, output):
        self.processor.process(namespace)
        output.send_user_message(aug_id, namespace['store'])

        if self.post_processor:
            output.flush_messages()
            self.post_processor.process(namespace)



class LogType:
    name_to_type = {}

    def __init__(self, name, log_type):
        self.log_type = log_type
        self.name = name
        LogType.name_to_type[name] = log_type

    @staticmethod
    def get_type(name):
        return LogType.name_to_type.get(name)


CURRENT_USER = LogType("currentUser", str)
ASSET_NAME = LogType("assetName", str)
ASSET_ID = LogType("assetId", str)
metric = LogType("metric", float)
ORG_ID = LogType("org_id", str)
REQUEST_URL = LogType("requestUrl", str)
REQUEST_ID = LogType("requestId", str)
AGENT_ID = LogType("agentId", str)
RULE = LogType("rule", dict)
WORKSPACE_ID = LogType("workspace", str)
REQUEST_REMOTE_IP = LogType("requestRemoteIp", str)
REQUEST_URI = LogType("requestUri", str)  # Used by tornado
STATUS_CODE = LogType("statusCode", str)
REQUEST_METHOD = LogType("requestMethod", str)



"""This module implements the Aug class."""
import time

from rook.logger import logger
from rook.user_warnings import UserWarnings
from rook.processor.error import Error
from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.stack_namespace import StackNamespace
from rook.processor.namespaces.python_utils_namespace import PythonUtilsNamespace
from rook.exceptions import RookRuleRateLimited


class Aug(object):
    """The Aug class is the skeleton that holds together all the components that define a modification to the application.

    This class brings together the following elements:
    - location - specifies when to run the modification.
    - extractor - specifies attributes to extract from the application's state before evaluating the modification.
    - condition - specifies an optional filter as to when to run the modification.
    - action - specifies the modification to preform.
    """

    def __init__(self, aug_id, location, extractor, condition, action, output, min_time_between_hits=0):
        """Build an Aug object from the individual elements."""
        self.aug_id = aug_id
        self._location = location
        self._extractor = extractor
        self._condition = condition
        self._action = action
        self._output = output
        self._last_executed = 0
        self._min_time_between_hits = min_time_between_hits
        self._enabled = True

        self._status = None

    def add_aug(self, trigger_services):
        """Use the location to add the Aug to the relevant trigger service."""
        try:
            self._location.add_aug(trigger_services, self._output, self)
        except Exception as exc:
            message = "Exception when adding aug"
            logger.exception(message)
            self.set_error(Error(exc=exc, message=message))

    def execute(self, frame_namespace, extracted):
        """Called by the trigger service to run the extractor, condition and action."""
        if not self._enabled:
            return

        try:
            with UserWarnings(self):
                logger.info("Executing aug-\t%s", self.aug_id)
                now = int(time.time() * 1000)

                if self._min_time_between_hits > 0 and now < self._last_executed + self._min_time_between_hits:
                    self.set_error(Error(exc=RookRuleRateLimited()))
                    self._enabled = False
                    return

                self._last_executed = now

                if self._extractor:
                    self._extractor.execute(frame_namespace, extracted)

                store = ContainerNamespace({})

                namespace = ContainerNamespace({
                    'frame': frame_namespace,
                    'stack': StackNamespace(frame_namespace),
                    'extracted': ContainerNamespace(extracted),
                    'store': store,
                    'temp': ContainerNamespace({}),
                    'python': PythonUtilsNamespace(),
                    'utils': PythonUtilsNamespace()
                })

                if not self._condition or self._condition.evaluate(namespace, extracted):
                    self._action.execute(self.aug_id, namespace, self._output)

        # Don't stop test exceptions from propagating
        except AssertionError:
            raise
        # Catch and silence everything else
        except Exception as exc:
            message = "Exception while processing Aug"
            logger.exception(message)
            self.send_warning(Error(exc=exc, message=message))

    def set_active(self):
        self._send_rule_status("Active")

    def set_pending(self):
        self._send_rule_status("Pending")

    def set_removed(self):
        self._send_rule_status("Deleted")

    def set_error(self, error):
        self._send_rule_status("Error", error)

    def set_unknown(self, error):
        self._send_rule_status("Unknown", error)

    def send_warning(self, error):
        self._output.send_warning(self.aug_id, error)

    def _send_rule_status(self, status, error=None):
        if self._status == status:
            return

        logger.info("Updating rule status for %s to %s", self.aug_id, status)

        self._status = status
        self._output.send_rule_status(self.aug_id, status, error)




import inspect
import six

from .namespace import Namespace
from .container_namespace import ContainerNamespace
from .python_object_namespace import PythonObjectNamespace

from rook.exceptions import RookAttributeNotFound


# Fixing a bug in PyPy
import platform
if 'PyPy' == platform.python_implementation():
    import sys
    import six
    FakeFrameType = type(next(six.itervalues(sys._current_frames())))
    from types import FrameType

    def isframe(object):
        return isinstance(object, (FrameType, FakeFrameType))

    inspect.isframe = isframe

class FrameNamespace(Namespace):

    def __init__(self, frame):
        super(FrameNamespace, self).__init__(self.METHODS)

        self.frame = frame
        self._filename, self._lineno, self._function, self._code_context, self._index = inspect.getframeinfo(frame, 0)

    def read_attribute(self, name):
        if name in self.frame.f_locals:
            return PythonObjectNamespace(self.frame.f_locals[name])
        elif name in self.frame.f_globals:
            return PythonObjectNamespace(self.frame.f_globals[name])
        else:
            raise RookAttributeNotFound(name)

    def filename(self, args=None):
        return PythonObjectNamespace(self._filename)

    def line(self, args=None):
        return PythonObjectNamespace(self._lineno)

    def function(self, args=None):
        return PythonObjectNamespace(self._function)

    def module(self, args=None):
        module = inspect.getmodule(self.frame)
        if module:
            return PythonObjectNamespace(module.__name__)
        else:
            return PythonObjectNamespace(None)

    def locals(self, args=None):
        if args:
            depth = int(args)
        else:
            depth = None

        result = {}
        for name, value in six.iteritems(self.frame.f_locals):
            ns = PythonObjectNamespace(value)

            if depth is not None:
                ns.dump_config.max_depth = depth
            elif name == 'self':
                ns.dump_config = PythonObjectNamespace.ObjectDumpConfig.conservative_limits(value)

            result[name] = ns

        return ContainerNamespace(result)

    def globals(self, args):
        return PythonObjectNamespace(self.frame.f_globals)

    def dump(self, args):
        return ContainerNamespace({
            'locals': self.locals(args),
            'module': self.module(),
            'filename': self.filename(),
            'line': self.line(),
            'function': self.function(),
        })

    def f_back(self):
        frame = self.frame.f_back

        if frame:
            return FrameNamespace(self.frame.f_back)
        else:
            return PythonObjectNamespace(None)

    def f_next(self):
        frame = self.frame.f_next

        if frame:
            return FrameNamespace(self.frame.f_next)
        else:
            return PythonObjectNamespace(None)

    METHODS = (filename, line, function, module, locals, globals, dump)



"""This module is in charge of managing the rook's module state.

The external interface for the module is Rook located in rook.rook."""

import os
import sys
import platform
import atexit

from rook.logger import logger

from rook.config import AgentAddress

from .augs.augs_manager import AugsManager
from .output import Output
from .com.agent_com import AgentCom
from .trigger_services import TriggerServices

from rook.exceptions import RookInterfaceException, RookVersionNotSupported

from .atfork import install_fork_handler


class _Singleton(object):
    """This is singleton is the class managing the module.

    It should never be referred to directly, instead use obj in this module."""

    def __init__(self):
        """Initialize the object, sets member variables."""
        self._check_version_supported()

        logger.info("Initializing Rook under process-%d", os.getpid())

        self._services_started = False

        self._trigger_services = TriggerServices()
        self._output = Output(self._trigger_services)
        self._aug_manager = None
        self._fork_handler = None

        self._agent_com = None

    def _start_trigger_services(self):
        """Start trigger services.

        Calling this method multiple times will have no effect.
        """
        # Don't double init services
        if self._services_started:
            return

        self._trigger_services.start()
        self._aug_manager = AugsManager(self._trigger_services, self._output)
        self._services_started = True

    def _stop_trigger_services(self):
        if not self._services_started:
            return

        self._aug_manager = None
        self._trigger_services.close()

        self._services_started = False

    def connect(self, token, host, port, tags=None):
        """Connect to the Agent."""
        if self._agent_com:
            raise RookInterfaceException("Multiple connection attempts not supported!")

        self._fork_handler = install_fork_handler()

        if not host:
            host = os.environ.get('ROOKOUT_AGENT_HOST', AgentAddress.HOST)

        if not port:
            port = int(os.environ.get('ROOKOUT_AGENT_PORT', AgentAddress.PORT))

        if not token:
            token = os.environ.get('ROOKOUT_TOKEN')

        self._start_trigger_services()

        if tags is not None:
            self._output.tags = tags

        logger.debug("Initiating AgentCom-\t%s:%d", host, int(port))
        self._agent_com = AgentCom(self._aug_manager, self._trigger_services, self._output, host, port, token)
        self._output.set_agent_com(self._agent_com)

        self._agent_com.connect_to_agent()
        logger.info("Successful connection to agent")

    def flush(self):
        self._output.flush_messages()

    def stop(self):
        logger.debug("Shutting down")

        if self._agent_com is not None:
            self._agent_com.stop()

        self._stop_trigger_services()

        self._agent_com = None

    @staticmethod
    def _check_version_supported():
        try:
            supported_platforms = ['pypy', 'cpython']
            supported_version = ['2.7', '3.5', '3.6', '3.7']

            current_platform = platform.python_implementation().lower()
            if current_platform not in supported_platforms:
                raise RookVersionNotSupported("Rook is not supported in this platform: " + current_platform)

            major, minor, _, _, _ = sys.version_info
            current_version = "{}.{}".format(major, minor)
            if current_version not in supported_version:
                raise RookVersionNotSupported("Rook is not supported in this python version: " + current_version)

        except Exception as e:
            import traceback
            traceback.print_exc()

            raise e


singleton_obj = _Singleton()


def exit_handler():
    try:
        logger.info("Exit handler called - flushing")
        if singleton_obj._agent_com:
            singleton_obj.flush()
    except:
        logger.exception("Flush failed")
    logger.info("Exit handler finished")


atexit.register(exit_handler)



import sys
import six
import os

from rook.logger import logger

if sys.platform in ('darwin', 'linux2', 'linux'):

    hook_installed = False
    original_os_fork = None

    import native_extensions

    def os_fork_hook():
        try:
            six.print_("[Rookout] Rookout does not support running in forking processes. Shutting down.")

            from .singleton import singleton_obj
            singleton_obj.stop()

            native_extensions.python_fork_handler_called = 1

        except:
            pass

        return original_os_fork()

    def install_fork_handler():
        global hook_installed
        if hook_installed:
            return
        hook_installed = True

        native_extensions.RegisterPreforkCallback()

        logger.debug("Installing os.fork handler")
        global original_os_fork
        original_os_fork = os.fork
        os.fork = os_fork_hook

else:
    def install_fork_handler():
        pass




from pyparsing import Combine, Optional, Word, Keyword, Literal, QuotedString, ZeroOrMore, oneOf, opAssoc, alphas, \
    alphanums, nums, infixNotation, ParserElement

from ...logger import logger

from rook.exceptions import RookInvalidArithmeticPath
from .basic_path import BasicPath
from ..namespaces.python_object_namespace import PythonObjectNamespace


# Some future ideas for this class:
# 1. Make this class self-contained without relying on BasicPath.
# 2. Add support for functions, lists and more...
# 3. Make this class the default class in some/all use cases.


class ArithmeticPath(object):

    NAME = 'calc'

    def __init__(self, configuration, factory):
        logger.info("Building Arithmetic Path")

        string_path = configuration['path']
        syntax = self._build_syntax()

        parsed = syntax.parseString(string_path, parseAll=True)
        if len(parsed) != 1:
            raise RookInvalidArithmeticPath(configuration)
        self._expr = parsed[0]

    def read_from(self, root_namespace):
        result = self._expr.eval(root_namespace)
        return PythonObjectNamespace(result)

    def write_to(self, root_namespace, value):
        raise NotImplementedError()

    @staticmethod
    def _build_syntax():

        class Operator(object):
            def eval(self, namespace):
                raise NotImplementedError()

            @staticmethod
            def _operatorOperands(token_list):
                """generator to extract operators and operands in pairs"""
                it = iter(token_list)
                while 1:
                    try:
                        yield (next(it), next(it))
                    except StopIteration:
                        break

        class EvalValue(Operator):
            def __init__(self, value):
                self._value = value

            def eval(self, namespace):
                return self._value

        class EvalPath(Operator):
            def __init__(self, tokens):
                self._path = BasicPath(tokens[0])

            try:
                PRIMITIVE_VALUES = (str, unicode, int, long, float, list)
            except NameError:
                PRIMITIVE_VALUES = (str, int, float, list)

            def eval(self, namespace):
                value = self._path.read_from(namespace).obj
                if isinstance(value, self.PRIMITIVE_VALUES):
                    return value
                else:
                    # TODO - this needs to be changed if we want to have arithmetic path as the default path
                    raise ValueError("Value is not supported by ArithmeticPath!", value)

        class EvalList(Operator):
            """Class to evaluate expressions building a list"""

            def __init__(self, tokens):
                self.tokens = tokens[1:-1:2]

            def eval(self, namespace):
                return [token.eval(namespace) for token in self.tokens]

        class EvalSignOp(Operator):
            """Class to evaluate expressions with a leading + or - sign"""

            def __init__(self, tokens):
                self.sign, self.value = tokens[0]

            def eval(self, namespace):
                mult = {'+': 1, '-': -1}[self.sign]
                return mult * self.value.eval(namespace)

        class EvalMultOp(Operator):
            """Class to evaluate multiplication and division expressions"""

            def __init__(self, tokens):
                self.value = tokens[0]

            def eval(self, namespace):
                result = self.value[0].eval(namespace)
                for op, val in self._operatorOperands(self.value[1:]):
                    if op == '*':
                        result *= val.eval(namespace)
                    elif op == '/':
                        result /= val.eval(namespace)
                return result

        class EvalAddOp(Operator):
            """Class to evaluate addition and subtraction expressions"""

            def __init__(self, tokens):
                self.value = tokens[0]

            def eval(self, namespace):
                result = self.value[0].eval(namespace)
                for op, val in self._operatorOperands(self.value[1:]):
                    if op == '+':
                        result += val.eval(namespace)
                    elif op == '-':
                        result -= val.eval(namespace)
                return result

        class AndOrOp(Operator):

            def __init__(self, tokens):
                self.value = tokens[0]

            def eval(self, namespace):
                result = bool(self.value[0].eval(namespace))
                for op, val in self._operatorOperands(self.value[1:]):
                    if op == 'and':
                        result = result and val.eval(namespace)
                    elif op == 'or':
                        result = result or val.eval(namespace)
                return result

        class EvalComparisonOp(Operator):
            "Class to evaluate comparison expressions"
            _opMap = {
                "<": lambda a, b: a < b,
                "<=": lambda a, b: a <= b,
                ">": lambda a, b: a > b,
                ">=": lambda a, b: a >= b,
                "!=": lambda a, b: a != b,
                "=": lambda a, b: a == b,
                "LT": lambda a, b: a < b,
                "LE": lambda a, b: a <= b,
                "GT": lambda a, b: a > b,
                "GE": lambda a, b: a >= b,
                "NE": lambda a, b: a != b,
                "EQ": lambda a, b: a == b,
                "<>": lambda a, b: a != b,
                "in": lambda a, b: a in b,
            }

            def __init__(self, tokens):
                self.value = tokens[0]

            def eval(self, namespace):
                val1 = self.value[0].eval(namespace)
                for op, val in self._operatorOperands(self.value[1:]):
                    fn = self._opMap[op]
                    val2 = val.eval(namespace)
                    if not fn(val1, val2):
                        break
                    val1 = val2
                else:
                    return True
                return False

        boolean = Keyword('True').setParseAction(lambda tokens: EvalValue(True)) | \
            Keyword('False').setParseAction(lambda tokens: EvalValue(False))
        integer = Combine(Optional(oneOf("+ -")) + Word(nums)).setName("integer") \
            .setParseAction(lambda tokens: EvalValue(int(tokens[0])))
        real = Combine(Optional(oneOf("+ -")) + Word(nums) + "." + Optional(Word(nums)) +
                       Optional(oneOf("e E") + Optional(oneOf("+ -")) + Word(nums))) \
            .setName("real").setParseAction(lambda tokens: EvalValue(float(tokens[0])))
        string = QuotedString('"').setName("string").setParseAction(lambda tokens: EvalValue(str(tokens[0])))

        values = boolean | string | real | integer

        namespace_path = Word(alphas, alphanums + ".[]'\"").setName("path").setParseAction(EvalPath)

        single_value = values | namespace_path
        list_value = (Literal("[") + ZeroOrMore(single_value + Literal(",")) + Optional(single_value) + Literal("]")). \
            setName("list").setParseAction(lambda tokens: EvalList(tokens))
        atom = single_value | list_value

        sign_op = oneOf('+ -')
        mult_op = oneOf('* /')
        plus_op = oneOf('+ -')
        and_op = Literal('and')
        or_op = Literal('or')
        comparison_op = oneOf("< <= > >= != = <>") ^ Keyword("LT") ^ Keyword("GT") ^ Keyword("LE") ^ \
                Keyword("GE") ^ Keyword("EQ") ^ Keyword("NE") ^ Keyword("in")

        expr = infixNotation(atom,
                             [
                                (sign_op, 1, opAssoc.RIGHT, EvalSignOp),
                                (mult_op, 2, opAssoc.LEFT, EvalMultOp),
                                (plus_op, 2, opAssoc.LEFT, EvalAddOp),
                                (comparison_op, 2, opAssoc.LEFT, EvalComparisonOp),
                                (and_op, 2, opAssoc.LEFT, AndOrOp),
                                (or_op, 2, opAssoc.LEFT, AndOrOp),
                            ])

        return expr



import inspect
import hashlib
import six
import subprocess

from rook.services.bdb_location_service import BdbLocationService
from rook.services.import_service import ImportService

from rook.exceptions import RookHashFailedException, RookHashMismatchException

from rook.logger import logger
from rook.processor.error import Error


class LocationFileLine(object):

    NAME = 'file_line'

    def __init__(self, arguments, processor_factory=None):
        self.filename = arguments.get('filename')
        # If we have a filename, make sure pyc is removed
        if self.filename:
            if isinstance(self.filename, str):
                # Some BDBs identify files by ID
                self.filename = self.filename.replace('.pyc', '.py')

        if self.filename is None:
            self.module_name = arguments['module_name'].lower()
        else:
            self.module_name = None

        self.lineno = arguments['lineno']

        self.file_hash = arguments.get('sha256')

        self.include_externals = arguments.get('includeExternals', False)

    def add_aug(self, trigger_services, output, aug):
        logger.info("Adding aug")

        def callback(module):
            try:
                self.test_hash(module)

                trigger_services.get_service(BdbLocationService.NAME).add_breakpoint_aug(module, self.lineno, aug)
            except Exception as exc:
                message = "Exception when adding aug"
                logger.exception(message)
                aug.set_error(Error(exc=exc, message=message))

        def removed():
            aug.set_removed()

        import_service = trigger_services.get_service(ImportService.NAME)
        if import_service.register_post_import_notification(aug.aug_id, self.module_name, self.filename, self.include_externals, callback, removed):
            aug.set_pending()

    def test_hash(self, module):
        # Don't test if we don't have a hash
        if not self.file_hash:
            return

        filepath = inspect.getsourcefile(module)
        if not filepath:
            raise RookHashFailedException(module.__name__)

        with open(filepath, 'rb') as f:
            if six.PY2:
                string = f.read().replace('\r\n', '\n').replace('\r\x00\n\x00', '\n\x00').replace('\r', '\n')
            else:
                string = f.read().decode().replace('\r\n', '\n').replace('\r\x00\n\x00', '\n\x00').\
                    replace('\r', '\n').encode('UTF8')

        hash = hashlib.sha256(string).hexdigest()
        if hash != self.file_hash:
            blob_hash = self.get_git_blob_hash(filepath)

            raise RookHashMismatchException(filepath, self.file_hash, hash, blob_hash)

    def get_git_blob_hash(self, file_path):
        return None




from .python_object_namespace import PythonObjectNamespace
from .dumped_object_namespace import DumpedObjectNamespace


class UnknownNamespace(DumpedObjectNamespace):

    def __init__(self, type, attributes={}):
        super(UnknownNamespace, self).__init__(type, "UnknownObject", attributes)

    def to_dict(self):
        return {
            u'@namespace': self.__class__.__name__,
            u'@common_type': self.common_type,
            u'@original_type': self.type,
            u'@attributes': self.get_attributes_dict()
        }

    def to_simple_dict(self):
        return u'<Unknown>'



import json
import threading
import time
import six
import sys
import traceback
import os
from six.moves.urllib.parse import urlparse


def enable_gevent_for_grpc():
    try:
        import gevent.monkey
    except ImportError:
        return

    if not gevent.monkey.is_module_patched('threading'):
        return

    import grpc.experimental.gevent as grpc_gevent
    grpc_gevent.init_gevent()


os.environ["GRPC_ENABLE_FORK_SUPPORT"] = "False"

enable_gevent_for_grpc()

from grpc import insecure_channel, RpcError, StatusCode, secure_channel,\
    ssl_channel_credentials

from rook.logger import logger

from ..services.bdb_location_service import BdbLocationService

from rook.protobuf import rook_pb2_grpc
from rook.protobuf import rook_pb2


from rook.config import AgentComConfiguration

from rook.exceptions import RookCommunicationException, RookInvalidToken


class AgentCom(object):
    """This class handles the communication itself."""

    def __init__(self, aug_manager, trigger_services, output, agent_host, agent_port, token):
        self._aug_manager = aug_manager
        self._output = output

        try:
            self._bdb = trigger_services.get_service(BdbLocationService.NAME)
        except:
            self._bdb = None

        self.agent_host = agent_host
        self.agent_port = agent_port
        self.token = token

        self._metadata = []

        if self.token:
            self._metadata.append(("rookout-token", self.token))

        self._channel = None
        self._client = None
        self._command_request = None
        self.request_timeout_in_seconds = None

        self._retries = 0
        self._last_successful_connection = 0
        self._backoff = AgentComConfiguration.BACK_OFF

        self._bg_connect_thread = None
        self._retry_thread_handle = None
        self._command_thread_handle = None

    def connect_to_agent(self):
        logger.debug("Attempting connection to agent")

        # Passing class as thread argument to retrieve return value, since in python 2.x
        #  closure variables are read-only
        class CommunicationThreadArguments(object):
            def __init__(self):
                self.t, self.v, self.bt = (None, None, None)
                self.success = False

        def bg_connect_to_agent(args):
            try:
                self._start_new_connnection()
                args.success = True
            except RookInvalidToken:
                args.t, args.v, args.bt = sys.exc_info()
            except RookCommunicationException:
                args.t, args.v, args.bt = sys.exc_info()
                logger.warn("Failed to connect to agent. Will keep trying")
                self._start_retry_thread()
            except Exception:
                args.t, args.v, args.bt = sys.exc_info()
                logger.error("Unexpected error: " + traceback.format_exc())

        arguments = CommunicationThreadArguments()
        self._bg_connect_thread = threading.Thread(target=bg_connect_to_agent, args=(arguments,))
        self._bg_connect_thread.daemon = True
        self._bg_connect_thread.start()
        self._bg_connect_thread.join(AgentComConfiguration.GRPC_TIMEOUT)

        # On windows OS first connection will most likely fail - but will succeed later on (in retry thread)
        if not arguments.success:
            if None not in [arguments.t, arguments.v, arguments.bt]:
                six.reraise(arguments.t, arguments.v, arguments.bt)
            raise RookCommunicationException()

    def send_rook_messages(self, messages):
        # timeout in seconds
        if self._client:
            self._client.send_messages(messages, metadata=self._metadata, timeout=self.request_timeout_in_seconds)

    def stop(self):
        self._output.flush_messages()
        self._output.stop_sending_messages()

        if self._command_request:
            self._command_request.cancel()
            self._command_request = None

        if self._client:
            self._client = None
            self._channel.close()
            self._channel = None

        if self._command_thread_handle:
            self._command_thread_handle.join()
        if self._bg_connect_thread:
            self._bg_connect_thread.join()
        if self._retry_thread_handle:
            self._retry_thread_handle.join()

    def _build_client(self):
        if self._client:
            self._client = None
            self._channel.close()
            self._channel = None

        keepaliveTime = 2 * 60 * 1000
        logger.debug("Connecting to agent-\t%s:%d", self.agent_host, int(self.agent_port))
        options = [('grpc.max_receive_message_length', AgentComConfiguration.MAX_MESSAGE_LENGTH),
                   ('grpc.keepalive_time_ms', keepaliveTime),
                   ('grpc.keepalive_timeout_ms', 2500),
                   ('grpc.http2.max_pings_without_data', 0),
                   # min_time_between_pings_ms by default is 5 mintues - needs to be less then keepaliveTime
                   ('grpc.http2.min_time_between_pings_ms', keepaliveTime - 1000)
                   ]

        self.request_timeout_in_seconds = AgentComConfiguration.REQUEST_TIMEOUT_SECS

        secure_connection = False
        new_host = self.agent_host

        parsed_url = urlparse(self.agent_host)
        if parsed_url.scheme != '':
            new_host = parsed_url.netloc
            secure_connection = (parsed_url.scheme == "https")

        connection_string = new_host + ':' + str(self.agent_port)
        if secure_connection:
            credentials = ssl_channel_credentials()
            self._channel = secure_channel(connection_string, credentials, options=options)
        else:
            self._channel = insecure_channel(connection_string, options=options)

        logger.debug("Creating client")
        self._client = rook_pb2_grpc.AgentManagementServiceStub(self._channel)

    def _start_new_connnection(self):
        logger.debug("Starting a new connection")

        self._build_client()

        logger.debug("Getting rook info")
        info = self._output.get_rook_info()

        logger.debug("Pinging")
        ping_message = rook_pb2.PingMessage()
        ping_message.value = 1
        try:
            self._client.ping(ping_message, timeout=AgentComConfiguration.GRPC_TIMEOUT, metadata=self._metadata)
        except Exception as e:
            if isinstance(e, RpcError) and e.code() in (StatusCode.CANCELLED, StatusCode.NOT_FOUND,
                                StatusCode.UNAUTHENTICATED, StatusCode.UNIMPLEMENTED,
                                StatusCode.PERMISSION_DENIED):
                    logger.error('Bad token while connecting to router', exc_info=1)
                    raise RookInvalidToken()
            message = "Failed to ping agent"
            logger.info(message, exc_info=1)
            six.raise_from(RookCommunicationException(message), e)

        logger.debug("Creating request")
        self._command_request = self._client.new_rook_notification(info, metadata=self._metadata)

        # Get initial augs
        success = self._command_loop(self._command_request)
        if not success:
            raise RookCommunicationException("Failed to connect to agent")

        self._last_successful_connection = int(time.time())

        # Setup update/retry channel in the background
        logger.debug("Creating commands thread")
        self._command_thread_handle = threading.Thread(target=self._command_thread, args=(self._command_request, ),
                                  name=AgentComConfiguration.COMMAND_THREAD_NAME)
        self._command_thread_handle.daemon = True
        self._command_thread_handle.start()

    def _command_loop(self, commands_request):
        logger.debug("Entering command loop")

        try:
            for command in commands_request:
                if command.command_type == rook_pb2.RookCommand.COMMAND_ADD_AUG:
                    logger.info("Got an add aug command")
                    try:
                        self._aug_manager.add_aug(json.loads(command.aug_json))
                    except:
                        logger.exception("Error processing new aug")

                elif command.command_type == rook_pb2.RookCommand.COMMAND_REMOVE_AUG:
                    logger.info("Got a remove aug command")
                    try:
                        self._aug_manager.remove_aug(command.aug_id)
                    except:
                        logger.exception("Error removing aug")

                elif command.command_type == rook_pb2.RookCommand.COMAND_INIT_FINISHED:
                    logger.info("Finished initialization")
                    self._output.start_sending_messages()
                    return "Init Finished"

                elif command.command_type == rook_pb2.RookCommand.COMMAND_CLEAR_AUGS:
                    logger.info("Got clear augs command")
                    try:
                        self._aug_manager.clear_augs()
                    except:
                        logger.exception("Error in clearing augs")

                elif command.command_type == rook_pb2.RookCommand.COMMAND_SET_ROOK_ID:
                    logger.info("Got set rook_id command- %s", command.rook_id)
                    self._output.set_rook_id(command.rook_id)

                else:
                    logger.error("Unknown command %s", command)
        except RpcError as exc:
            # This exception can occur during process shutdown
            if StatusCode.CANCELLED == exc.code():
                self._output.stop_sending_messages()
                return "Process Exiting"

            if self._retries:
                logger.info("Error in command loop", exc_info=1)
            else:
                logger.exception("Error in command loop")
        except:
            if self._retries:
                logger.info("Error in command loop", exc_info=1)
            else:
                logger.exception("Error in command loop")

        self._output.stop_sending_messages()
        logger.warn("Exiting with failure")
        return None

    def _command_thread(self, command_request):
        try:
            logger.debug("Command thread started")
        except TypeError:
            # When running python -m rook, there is a race condition where the thread might start
            # during interpreter shut down
            return

        # Disable bdb monitoring of this thread
        if self._bdb:
            self._bdb.ignore_current_thread()

        if command_request:
            ret = self._command_loop(command_request)
            # True here means the process is exiting
            if ret:
                return

        self._start_retry_thread()

    def _start_retry_thread(self):
        if not self._retry_thread_handle or not self._retry_thread_handle.is_alive():
            logger.debug("Starting retry thread")
            self._retry_thread_handle = threading.Thread(target=self._retry_thread,
                                                         name=AgentComConfiguration.COMMAND_THREAD_NAME)
            self._retry_thread_handle.daemon = True
            self._retry_thread_handle.start()
        else:
            logger.warn("Retry thread already running")

    def _retry_thread(self):
        logger.info("Starting retry thread")

        # Disable bdb monitoring of this thread
        if self._bdb:
            self._bdb.ignore_current_thread()

        min_time_since_last_successful_connection = 60
        if int(time.time() > self._last_successful_connection + min_time_since_last_successful_connection):
            self._retries = 0
            self._backoff = AgentComConfiguration.BACK_OFF

        while True:
            try:
                if not self._client:
                    break

                self._start_new_connnection()
                break
            except:
                # Interpreter shutdown
                if not logger:
                    return

                if self._retries > 0:
                    self._backoff = min(self._backoff * 2, AgentComConfiguration.MAX_SLEEP)
                    logger.debug("Sleeping for %d seconds, retry num %d", self._backoff, self._retries)
                    time.sleep(self._backoff)

                self._retries += 1



from rook.exceptions import RookWriteAttributeNotSupported, RookAttributeNotFound, RookMethodNotFound


class Namespace(object):
    def __init__(self, methods=None):
        self._methods = {}

        if methods:
            for method in methods:
                self._methods[method.__name__] = method

    def call_method(self, name, args):
        try:
            method = self._methods[name]
        except KeyError:
            raise RookMethodNotFound(type(self), name)

        return method(self, args)

    def read_attribute(self, name):
        try:
            return getattr(self, name)
        except AttributeError:
            raise RookAttributeNotFound(name)

    def write_attribute(self, name, value):
        raise RookWriteAttributeNotSupported(type(self), name)

    def read_key(self, key):
        try:
            return self[key]
        except KeyError:
            raise RookAttributeNotFound(key)

    def to_dict(self):
        raise NotImplementedError()

    def to_simple_dict(self):
        raise NotImplementedError()



"""This module handles all the output generated by the Rook.

The output generated by Rook into the following types of messages:
- Rook information describing the rook and the application, usually at startup.
- User messages describing the application's behaviour.
- Log messages describing the rook's behaviour.

Output (except rook information) is batched and forwarded asynchronously to the Agent."""

import sys
import os
import threading
import time
import logging
import six
import traceback

import platform
# calling platform.platform to avoid dead-lock on secondary thread (only Windows OS)
platform.platform()

from rook.config import OutputConfiguration, VersionConfiguration

from .services.bdb_location_service import BdbLocationService

from rook.logger import logger
from rook.logger.namespace_serializer_handler import NamespaceSerializerHandler

from rook import git

from rook.machine import add_info

from rook.processor.namespace_serializer import NamespaceSerializer

from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.python_object_namespace import PythonObjectNamespace

from rook.protobuf import rook_pb2


class Output(object):
    """The Output class manages all Rook's output.

    The object is initialized to a passive state, collecting all the data written to it.
    Only after the object is given a connection to the agent does message sending begin."""

    def __init__(self, trigger_services):
        """Build the output object in passive state."""
        self._trigger_services = trigger_services
        self._rook_id = ''

        self._lock = threading.RLock()
        self._messages = rook_pb2.RookMessages()

        self._agent_com = None

        self._closing = False
        self._thread = None
        self._thread_lock = threading.RLock()

        self._log_limit_reached = False
        self._message_limit_reached = False
        self._status_update_limit_reached = False

        NamespaceSerializerHandler.output = self

        self.tags = []

    def close(self):
        NamespaceSerializerHandler.output = None

        self._stop_thread()

    def set_rook_id(self, rook_id):
        self._rook_id = rook_id

    def set_agent_com(self, agent_com):
        self._agent_com = agent_com

    def start_sending_messages(self):
        """Start the background thread sending data to agent"""
        self._start_thread()

    def stop_sending_messages(self):
        self._stop_thread()

    def get_rook_info(self):
        message = rook_pb2.RookInformation()
        message.rook_id = NamespaceSerializer.normalize_string(self._rook_id)

        # Everything else is basically optional and so we ignore errors in it
        try:
            message.version = NamespaceSerializer.normalize_string(VersionConfiguration.VERSION)
            message.commit = NamespaceSerializer.normalize_string(VersionConfiguration.COMMIT)

            message.platform = "python"
            message.platform_type = NamespaceSerializer.normalize_string(platform.python_implementation())
            message.platform_version = NamespaceSerializer.normalize_string(sys.version)
            message.executable = NamespaceSerializer.normalize_string(sys.argv[0])

            for arg in sys.argv:
                message.command_arguments.append(NamespaceSerializer.normalize_string(arg))

            message.tags.extend(self.tags)

            # Try to get the proper git tag
            try:
                user_commit = os.environ.get('ROOKOUT_COMMIT')
                if not user_commit:
                    git_root = os.environ.get('ROOKOUT_GIT')
                    if not git_root:
                        git_root = git.find_root(os.path.dirname(os.path.abspath(sys.argv[0])))
                    if git_root:
                        user_commit = git.get_revision(git_root)
            except:
                logger.exception("Failed to get git tag")
                user_commit = ""

            if user_commit:
                message.user_commit = NamespaceSerializer.normalize_string(user_commit)
            else:
                message.user_commit = ""

            message.process_id = str(os.getpid())

            add_info(message)
        except:
            logger.exception("Unexpected exception when building rook info")

        return message

    def send_warning(self, rule_id, error):
        self.send_rule_status(rule_id, "Warning", error)

    def send_rule_status(self, rule_id, active, error):
        try:
            if len(self._messages.rule_status_updates) > OutputConfiguration.MAX_STATUS_UPDATES:
                if not self._status_update_limit_reached:
                    logger.error("Limit reached (%d), dropping status updates", OutputConfiguration.MAX_STATUS_UPDATES)
                    self._message_limit_reached = True
                return

            rule_status = self._messages.rule_status_updates.add()
            rule_status.rook_id = NamespaceSerializer.normalize_string(self._rook_id)
            rule_status.rule_id = NamespaceSerializer.normalize_string(rule_id)
            rule_status.active = NamespaceSerializer.normalize_string(active)
            rule_status.time.GetCurrentTime()

            if error:
                error.dump(rule_status.error)
        except:
            logger.exception("Failed to report rule status")

    def send_user_message(self, aug_id, arguments):
        with self._lock:
            if len(self._messages.user_messages) > OutputConfiguration.MAX_ITEMS:
                if not self._message_limit_reached:
                    logger.error("Limit reached (%d), dropping user messages", OutputConfiguration.MAX_ITEMS)
                    self._message_limit_reached = True
                return

            message = self._messages.user_messages.add()
            message.aug_id = NamespaceSerializer.normalize_string(aug_id)
            message.time.GetCurrentTime()

            if arguments and arguments.size(''):
                NamespaceSerializer().dump(arguments, message.arguments)

    def send_log_message(self, level, time, filename, lineno, text, formatted_message, arguments):
        with self._lock:
            if len(self._messages.log_messages) >= OutputConfiguration.MAX_LOG_ITEMS:
                if not self._log_limit_reached:
                    self._log_limit_reached = True

                    try:
                        # We cannot preform a recursive call, so we override original message
                        level = logging.ERROR
                        filename = __file__
                        lineno = 0
                        text = "Limit reached (%d), dropping log messages"
                        formatted_message = text
                        arguments = ContainerNamespace({"args": PythonObjectNamespace(OutputConfiguration.MAX_ITEMS)})
                    except:
                        six.print_("[Rookout] Error when handling output queue size limit", file=sys.stderr)
                        traceback.print_exc()
                        return
                else:
                    return

            message = self._messages.log_messages.add()
            message.level = int(level)
            message.time.FromMilliseconds(int(time * 1000))
            message.filename = NamespaceSerializer.normalize_string(filename)
            message.lineno = int(lineno)
            message.text = NamespaceSerializer.normalize_string(text)
            message.formatted_message = NamespaceSerializer.normalize_string(formatted_message)
            if arguments and arguments.size(''):
                NamespaceSerializer().dump(arguments, message.arguments, log_errors=False)

    def flush_messages(self):
        # Copy messages aside while holding the lock
        with self._lock:
            messages = self._messages
            self._messages = rook_pb2.RookMessages()

        # If we got any messages, send them
        if messages.user_messages or messages.log_messages or messages.rule_status_updates:
            try:
                messages.rook_id = self._rook_id
                self._agent_com.send_rook_messages(messages)

                self._log_limit_reached = False
                self._message_limit_reached = False
                self._status_update_limit_reached = False
            except:
                # Logger might not exist in interpreter shutdown
                if logger:
                    logger.exception("Sending messages to agent failed")

    def _start_thread(self):
        with self._thread_lock:
            if self._thread:
                return

            self._closing = False
            self._thread = threading.Thread(target=self._output_thread, name=OutputConfiguration.THREAD_NAME)
            self._thread.daemon = True
            self._thread.start()

    def _stop_thread(self):
        # If lock is already dead, just return
        if not self._thread_lock:
            return

        with self._thread_lock:
            if self._thread:
                self._closing = True

                # If threading was monkey patched by gevent waiting on thread will likely throw an exception
                try:
                    from gevent.monkey import is_module_patched
                    if is_module_patched("threading"):
                        time.sleep(OutputConfiguration.FLUSH_TIME_INTERVAL)
                except:
                    pass

                self._thread.join()
                self._thread = None

    def _output_thread(self):
        """This is the thread communication sending the batched messages to the agent."""

        # Don't allow this thread to be monitored
        try:
            self._trigger_services.get_service(BdbLocationService.NAME).ignore_current_thread()
        except KeyError:
            pass

        # Until we are asked to terminate
        while not self._closing:
            # Flush messages
            self.flush_messages()

            # If interpreter is not in shutdown
            if time:
                # Wait for next check
                time.sleep(OutputConfiguration.FLUSH_TIME_INTERVAL)
            else:
                break



import os


# note: the logger is writing to stack driver -> we read from stack driver
# stackdriver -> pubsub (google service) -> fluentd from pubsub
def add_backend_specific_logs(logging_values):
    if os.environ.get('SYSTEM_APP') == "rookout_backend":
        import flask
        if flask.has_request_context():
            request_url, request_id, agentId = flask.request.url, flask.request.headers.get("RUID"), flask.request.headers.get("X-Agent-ID")
            logging_values["requestUrl"] = request_url
            logging_values["requestId"] = request_id
            logging_values["agentId"] = agentId
            logging_values["requestRemoteIp"] = flask.request.remote_addr



from rook.services.logging_location_service import LoggingLocationService


class LocationLogHandler(object):

    NAME = 'log_handler'

    def __init__(self, arguments, processor_factory):
        self.logger = arguments['logger']

    def add_aug(self, trigger_services, output, aug):
        trigger_services.get_service(LoggingLocationService.NAME).add_logging_aug(self.logger, aug)



"""This module initializes the Rook using default settings when imported."""
import os


def _auto_start():
    disable = os.environ.get('ROOKOUT_DISABLE_AUTOSTART', None)

    if disable is not None and disable != '0':
        return

    import rook
    rook.start()


_auto_start()
